/*
main difference to QUICK GRN: there can be any number of cis sites for a gene and 
cis sites themselves can be complex, i.e. consist of several protein binding sites.
this of course makes a different seperation of genome parts necessary as well as a
different mode of creating offspring.
cis sites are delimited by "2", every gene on the genome is delimited by "3"!

changes mainly made to the constructor, look for: >>process "genes"<<
and the getOffspring method
*/

/*
GRN algorithm in an artificial evolution framework to model biological clocks
Copyright (C) 2005  Johannes Knabe
j.f.knabe@herts.ac.uk

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

import java.util.*;
import java.io.*;
 
public class KnabeGRN implements individual { 

public PrintWriter dynWriter=null;//this might be set from the outside; if !=null protein concentrations will
                                  //be written to it, so you can keep track of the dynamics - close it from the outside!!!
public static int inputProtein=1; //this protein's level will be increased depending on "sensory" input
public static int inputProtein2=3;
public static int inputProtein3=4;
public static int outputProtein=6; //this protein's level will determine the system#s output
//inoutScaling: Input and Output are between 0 and 1, so we do some
//scaling when inputting/outputting protein concentrations
//150 was chosen because it is the ceiling of on genes output, i.e.
//one gene can produce maximally 150 instances of a protein per cycle
public static double inoutScaling=150.0;
public static double outval; //just caches the last output of our net

public double mutationProb=0.01;
public double crossoverProb=0.9;
public int howLong=0; //lifetime counter
public static int lifetime=400;
public static int elitism=1;
public static int compPool=15;
public static int geneNo=0;    //number of genes, better change from outside!
//public int cisPerGene;//cis docking sites for each gene
public static int wavelength=20;
public static double noiseLevel=0.1;//when noise is applied to input this gives the std deviation of the gaussion noise
public static int blackoutLength  =(int)(0.05*lifetime);//how long should a blackout last - e.g. 10% of wave or life
public static int blackoutInterval=(int)(lifetime/2);   //how often should a blackout occur - e.g. once in a lifetime or once every wavelength?
public static double veryBadFitness=9999.0; //always use this if you want a fitness value so bad that it will be beaten by every individual
public static double[] decayRates={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0,1.0,1.0,1.0,1.0,1.0,1.0}; //\in [0..1]
public static double[] saturationValues={10,25,50,100,150,200,300,500};
public static double[] bindingProportions={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0,1.0,1.0,1.0,1.0,1.0,1.0}; //\in [0..1]
public static int whichInputFunction=0;
public static int whichFitnessMeasure=0;
public input myInOut=null;

public static void setInputFunction(int to){
whichInputFunction=to;
}

public static void setFitFunction(int to){
whichFitnessMeasure=to;
}

/*returns input function */
public input getIn(){
switch(whichInputFunction){
case 0 : return new sigmInp(wavelength);        //sinusoidal
case 1 : return new posSigInp(wavelength);      //positive sinusoidal
case 2 : return new stepInp(wavelength);        //step function
case 3 : return new pulseInp(wavelength);       //pulse function
case 4 : return new sigmInp(wavelength,noiseLevel);//sinusoidal with gaussian noise, param: std. dev.
case 5 : return new posSigInp(wavelength,noiseLevel);//positive sinusoidal with gaussian noise, param: std. dev.
case 6 : return new stepInp(wavelength,noiseLevel);//step function with gaussian noise, param: std. dev.
case 7 : return new pulseInp(wavelength,noiseLevel);//pulse function with gaussian noise, param: std. dev.
case 8 : return new sigmInpBlackout(wavelength,blackoutLength,blackoutInterval);  //sinusoidal with blackout periods
case 9 : return new posSigInpBlackout(wavelength,blackoutLength,blackoutInterval);//positive sinusoidal with blackout periods
case 10: return new stepInpBlackout(wavelength,blackoutLength,blackoutInterval);  //step function with blackout periods
case 11: return new pulseInpBlackout(wavelength,blackoutLength,blackoutInterval); //pulse function with blackout periods
case 12 : return new sigmInpBlackout(wavelength,noiseLevel,blackoutLength,blackoutInterval);  //sinusoidal with blackout periods
case 13 : return new posSigInpBlackout(wavelength,noiseLevel,blackoutLength,blackoutInterval);//positive sinusoidal with blackout periods
case 14: return new stepInpBlackout(wavelength,noiseLevel,blackoutLength,blackoutInterval);  //step function with blackout periods
case 15: return new pulseInpBlackout(wavelength,noiseLevel,blackoutLength,blackoutInterval); //pulse function with blackout periods
case 50: return new inverseSigmInp(wavelength); //sinusoidal, but will have a maximum when sigmInp has a minimum
default:System.out.println("ERROR NO SUcH INPUT");  System.exit(1);
}
return null;
}

/*returns fitness function*/
public fitness getFit(){
switch(whichFitnessMeasure){
 case 0 : return new sigmFit(wavelength); //sinusoidal output is desired
 case 1 : return new stepFit(wavelength); //step function output is desired
default:System.out.println("Error: No such fitness function!");  System.exit(1);
}
return null;
}

public double fitness;
public double bindingProportion;
public double saturationValue;
public double[] proteinDecay;
public int[] cisSitesReceptiveFor; //number of cis-sites that match a protein (sum over all genes and cis-sites)
public int[] genome;
public gene[] genes;
public double[] concentrations;
public int[] geneOutputProtein;
public int[] geneOutputActivity;
//public int[][] geneCisReceptor;
//public int[][] geneCisReceptorInhib;
public boolean debug;

public static int fromBinary(int[] bin){
int res=0;
for(int i=0;i<bin.length;i++){
  res=res<<1;
  res+=bin[i];
  }
return res;
}

/*returns a 3 "bit" array always*/
public static int[] toBinary(int bin){
int[] res=new int[3];
for(int i=2;i>=0;i--){
  res[i]=bin % 2;
  bin/=2;
}  
return res;
}

//int array to string conversion ({1,0,1}->"101")
public static String intArrToString(int[] in){
String ret="";
for(int i=0;i<in.length;i++)
  ret=ret+in[i];
return ret;
}

//string to int array conversion ("101"->{1,0,1})
public static int[] stringToIntArr(String in){
int[] ret=new int[in.length()];
for(int i=0;i<ret.length;i++)
  ret[i]=Integer.parseInt(""+in.charAt(i));
return ret;
}

public KnabeGRN(int[] geno){
debug=false;
fitness=veryBadFitness; /*small fitness is good, so this is pretty bad*/
genome=new int[geno.length];
for(int i=0;i<geno.length;i++)
  genome[i]=geno[i];
concentrations=new double[8]; //eight proteins are used
//process genome
proteinDecay=new double[8];
geneOutputProtein=new int[geneNo];
geneOutputActivity=new int[geneNo]; //set activity type (default 'on' / default 'off')
//geneCisReceptor=new int[geneNo][cisPerGene];
                                    //site is inhibitory if ==1 or activatory, else
//geneCisReceptorInhib=new int[geneNo][cisPerGene];
int offset=0;
//process "genes"
String[] splitGeno=(intArrToString(genome)).split(""+3); //3 is the delimiter of genes
genes=new gene[geneNo];
int[] currGene;
int count=0;
for(int i=0;i<splitGeno.length-1;i++){
  if(splitGeno[i].length()>3)//needs to code for a protein at least to qualify as a gene
    genes[count++]=new KnabeGene(stringToIntArr(splitGeno[i]),this);
}

cisSitesReceptiveFor=new int[8];
for(int j=0;j<8;j++){   //for every protein
  for(int i=0;i<geneNo;i++){ //for every gene
   if(genes[i]!=null)
    //sum up the receptive cis-sites
    cisSitesReceptiveFor[j]+=genes[i].getNoOfCisSitesForProt(j);
  }
}
//genes are done, now decode the "global" variables of this genome
offset=genome.length-39;
int[] dec=new int[4];
for(int i=0;i<8;i++){
  System.arraycopy(genome,offset,dec,0,4);
  proteinDecay[i]=lookupDecayRates(dec);
  offset+=4;
}

//TODO get original values for the lookup table
int[] bind=new int[4];
System.arraycopy(genome,offset,bind,0,4);
bindingProportion=lookupBindingProp(bind);

offset+=4;

int[] sat=new int[3];
System.arraycopy(genome,offset,sat,0,3);
saturationValue=lookupSaturationValue(sat);
}


public static double lookupDecayRates(int[] in){
int what=fromBinary(in);
return decayRates[what];
}

public static double lookupBindingProp(int[] in){
int what=fromBinary(in);
return bindingProportions[what];
}

public static double lookupSaturationValue(int[] in){
int what=fromBinary(in);
return saturationValues[what];
}

public void applySaturationCheck(){
//make sure that every protein is not more present than it's
//saturationValue allows! (important: do we call this before decay
//or after decay? Arbitraryly decided for before)
for(int i=0;i<concentrations.length;i++){
  if(concentrations[i]>saturationValue)
    concentrations[i]=saturationValue;
  }
}

public double process(double in, double in2, double in3){
double ret=process(in);
concentrations[inputProtein2]+=in2*inoutScaling;
concentrations[inputProtein3]+=in3*inoutScaling;
return ret;
}

public double process(double in){
double res=.0;
/*was used for debug
for(int i=0;i<concentrations.length;i++)
  System.out.print(concentrations[i]+",");
System.out.println("");*/
//1. bind unbound proteins to matching cis sites
//concentrations/cisSitesReceptiveFor := per site fraction
for(int i=0;i<concentrations.length;i++){
  for(int j=0;j<genes.length;j++){
   if(genes[j]!=null){
    if(cisSitesReceptiveFor[i]!=0)
     //distribute every cis-site it's share
     genes[j].bindingProtein(i,((double)concentrations[i]*bindingProportion)/((double)cisSitesReceptiveFor[i]));
    else //avoid NaN
     genes[j].bindingProtein(i,((double)concentrations[i]*bindingProportion));
   }
  }
  //of course this is lost now:
  concentrations[i]-=((double)concentrations[i]*bindingProportion);
}
//2. calculate activation of genes and give result to genes output function
for(int i=0;i<concentrations.length;i++){
  for(int j=0;j<genes.length;j++){
   if(genes[j]!=null)
    concentrations[i]+=genes[j].getOutputForProtein(i);
    }
  }
applySaturationCheck();
//3. decay according to protein specific decay rate
for(int i=0;i<concentrations.length;i++)
  concentrations[i]=concentrations[i]*proteinDecay[i];
//4. output and input
res=concentrations[outputProtein]/inoutScaling;
concentrations[inputProtein]+=in*inoutScaling;
applySaturationCheck();
/*was used for debug
for(int i=0;i<concentrations.length;i++)
  System.out.print(concentrations[i]+",");
System.out.println("");*/
if(dynWriter!=null){
  dynWriter.print(howLong+" ");
  for(int i=0;i<concentrations.length;i++)
    dynWriter.print(concentrations[i]+" ");
  dynWriter.println();
}
if(Double.isNaN(res)) //was used to punish NaN systems, which now hopefully 
  res=99.0;           //nevers happens anyway
outval=res;
return res;
}

public void live2(input in, fitness fit, int length){
//FOR TESTING ONLY (drawing gnuplot curves):
debug=true;
double inval;
double fitval;
for(int i=0;i<length;i++){
  inval=in.getVal();
  outval=process(inval);
  fitval=fit.procOutput(outval);
  System.out.println(i+" "+inval+" "+outval+" "+fitval);
  }
}

public void live3(input in, fitness fit, int length, PrintWriter out){
//FOR TESTING ONLY (drawing gnuplot curves):
debug=true;
double inval;
double fitval;
for(int i=0;i<length;i++){
  inval=in.getVal();
  outval=process(inval);
  fitval=fit.procOutput(outval);
  out.println(howLong+"  "+outval);
  howLong++;
  //out.println(i+" "+inval+" "+outval+" "+fitval);
  }
}

public void live(input in, fitness fit, int length, double toBeat){
//original
///*
boolean didNotBeat=false;
for(int i=0;i<length;i++){
  fit.procOutput(process(in.getVal()));
  if((length % 20 ==0)&&(fit.getFitness()>toBeat)){
    didNotBeat=true;
    break;
    }
  }
if(didNotBeat)
  fitness=veryBadFitness;
else
  fitness=fit.getFitness();
/*//FOR TESTING ONLY (no input, no output):
concentrations[1]=130;
for(int i=0;i<length;i++){
  System.out.println(i+" "+concentrations[1]+" "+concentrations[6]);
  process(0.0);
  }*/
}

public void live(input in, fitness fit, int length){
//original
///*
for(int i=0;i<length;i++){
  outval=process(in.getVal());
  fit.procOutput(outval);
  }
fitness=fit.getFitness();
/*//FOR TESTING ONLY (no input, no output):
concentrations[1]=130;
for(int i=0;i<length;i++){
  System.out.println(i+" "+concentrations[1]+" "+concentrations[6]);
  process(0.0);
  }*/
}

public void liveOutputGeneActivity(input in, fitness fit, int length, PrintWriter geneActivity){
//add to KnabeGene:
//public double lastRet; //at the start of file
//lastRet=ret; //last line of the function getOutputForProtein
geneActivity.println("#gene activity data from the genome:\n#"+intArrToString(genome)+"\n#having "+genes.length+"genes");
for(int i=0;i<length;i++){
  //normal processing
  fit.procOutput(process(in.getVal()));
  //special gene activity output
  for(int j=0;j<genes.length;j++)
   if(genes[j]!=null)
    geneActivity.print((int)((KnabeGene)genes[j]).lastRet+" ");
  geneActivity.println();
  }
fitness=fit.getFitness();
geneActivity.println("#overall fitness: "+fitness);
}

public double getFitness(){
return fitness;
}

public int[] getGenome(){
return genome;
}

public individual getRandomInd(){
//we start always very simple - one cis site with one binding protein per gene
int[] gen=new int[KnabeGRN.geneNo*10+39];//12:4 for a cis site + 2 for cis delimiters + 4 for protein produced + 2 for gene delimiters
java.util.Random rand=new java.util.Random();
int next;
for(int i=0;i<gen.length;i++){
  next=rand.nextInt(2);
  gen[i]=next;
}
for(int i=0;i<KnabeGRN.geneNo;i++){
  gen[i*10+9]=3;
  gen[i*10+4]=2;
}
return new KnabeGRN(gen);
}

/*returns lifetime (length for live method)*/
public int getLifetime(){
return lifetime;
}

/*returns elitism (number of individuals carried over [copied] to next gen)*/
public int getElitism(){
return elitism;
}
public int getCompPool(){
return compPool;
}

/*returns offspring generated from input individuals*/
//HERE: from two parents two offspring are generated
public individual[] getOffspring(individual[] parents){
KnabeGRN[] ret=new KnabeGRN[2];
int[] par1=parents[0].getGenome();
int[] par2=parents[1].getGenome();
java.util.Random rand=new java.util.Random();
String p1="";
String p2="";
String f1="";
String f2="";
p1=intArrToString(par1);
p2=intArrToString(par2);
//begin copying and maybe crossover

if(rand.nextDouble()>=crossoverProb){//no crossover
f1=p1;
f2=p2;
}
else{                                //do crossover
  //a crossover should always occur within the same compartment, i.e.
  //in the gene of the same number (although offset position might vary) or
  //in the part coding for global variables (offset may not vary - always same size)
  String[] splitGeno1=p1.split(""+3);
  String[] splitGeno2=p2.split(""+3);
  int crossoverComp=rand.nextInt(splitGeno1.length);
  if(splitGeno1[crossoverComp].length()<2) //should never happen....
    crossoverComp++;
  int offset;
 try{
  for(int i=0;i<splitGeno1.length;i++){
    if(crossoverComp==i){ //here it happens
      int crossoverPoint;
      if(splitGeno1[i].length()==0)
        crossoverPoint=0;
      else
        crossoverPoint=rand.nextInt(splitGeno1[i].length());
      if(crossoverComp==splitGeno1.length-1)
        offset=0;
      else
        offset=(int)(rand.nextGaussian()*4); //here it becomes interesting: the genome might become longer or shorter
      //do it for f1
      f1+=splitGeno1[i].substring(0,crossoverPoint);
      if(crossoverPoint+offset<0) //prevent IOErrors!
        offset=-crossoverPoint;
      if(!(crossoverPoint+offset>splitGeno2[i].length())) //prevent IOErrors!
        f1+=splitGeno2[i].substring(crossoverPoint+offset);

      //while(f1.length()-f1.lastIndexOf(""+3)<9) //prevent error: minimum length is 4
      //  f1+=rand.nextInt(2);

      //do it for f2
      if(crossoverPoint>splitGeno2[i].length()) //of course this might be shorter then splitGeno1[i], which is used for determining "crossoverPoint"
        crossoverPoint=splitGeno2[i].length();
      f2+=splitGeno2[i].substring(0,crossoverPoint);
      if(crossoverPoint-offset<0) //prevent IOErrors!
        offset=crossoverPoint;
      if(!(crossoverPoint-offset>splitGeno1[i].length())) //prevent IOErrors!
        f2+=splitGeno1[i].substring(crossoverPoint-offset);

      //while(f2.length()-f2.lastIndexOf(""+3)<9) //prevent error: minimum length is 4
      //  f2+=rand.nextInt(2);
    }
    else
      if(i<crossoverComp){
        f1+=splitGeno1[i];
        f2+=splitGeno2[i];
        }
      else{
        f1+=splitGeno2[i];
        f2+=splitGeno1[i];
        }
  if(i<splitGeno1.length-1){ //number of compartments stays the same always: number of genes plus the part coding for global variables
    f1+="3";
    f2+="3";
    }
  }//end for
 }//end try
 catch(Exception e){
  f1=p1;
  f2=p2;
  System.err.println("Error when crossing over the following genomes:\n"+p1+"\n\n"+p2+"\n\n"+e);
 }
}
//end copying and maybe crossover
int[] gene1=stringToIntArr(f1);
int[] gene2=stringToIntArr(f2);
//begin maybe mutation
for(int i=0;i<gene1.length;i++){
  if((gene1[i]<2)&&(rand.nextDouble()<=mutationProb)){
    if(gene1[i]==0)
      gene1[i]=1;
    else
      gene1[i]=0;
    }
  }
for(int i=0;i<gene2.length;i++){
  if((gene2[i]<2)&&(rand.nextDouble()<=mutationProb)){
    if(gene2[i]==0)
      gene2[i]=1;
    else
      gene2[i]=0;
    }
  }
//end maybe mutation
ret[0]=new KnabeGRN(gene1);
ret[1]=new KnabeGRN(gene2);
return ret;
}

public void toFile(String name){
File file = null;
PrintWriter out = null;
try{
  file = new File(name+".txt");
  out = new PrintWriter(
	new OutputStreamWriter(
	new FileOutputStream(file)));
}
catch(Exception e){
  System.err.println("Error creating output file!\n"+e);
}
out.println("#fitness: "+getFitness());
out.print("{");
for(int i=0;i<genome.length-1;i++){
  out.print(genome[i]+",");
}
out.println(genome[genome.length-1]+"}");
out.close();
}

public String toString(){
String ret="";
ret+="\ngenes: ";
for(int i=0;i<genome.length;i++){
ret+=genome[i]+",";
}
ret+="\nFitness: "+fitness+"\n";
return ret;
}

}